# -*- coding: utf-8 -*

#  =============================================================================
# Aeronautics Institute of Technologies (ITA) - Brazil
# University of Coimbra (UC) - Portugal
# Arthur Dantas Mangussi - mangussiarthur@gmail.com
# =============================================================================

import matplotlib.pyplot as plt
import missingno as msno
import pandas as pd
from typing import Tuple,Optional
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import random as rng

class PlotMissingData:
  """
  A class to visualize missing data in a DataFrame using different types of plots.

  Attributes
  ----------
  data_missing : pd.DataFrame
      The DataFrame containing the data with potential missing values.
  original : pd.DataFrame
      The complete DataFrame.
  palette : str
      The string to indicate the color palette of the plot.
  custom_color : tuple of float or None
      The custom color in RGB format, e.g., (1, 1, 1), or None if not specified.

  Methods
  -------
  visualize_miss(visualization_type: str = "normal", save: bool = True, path_save_fig: str = None)
      Visualizes the missing data using the specified type of plot and optionally saves the plot.
  """
  def __init__(self,
               data_missing:pd.DataFrame,
               data_original:pd.DataFrame,
               palette:str="gray",
               custom_color:Optional[Tuple[float, float, float]]=None) -> None:

    if not isinstance(data_missing, pd.DataFrame):
            raise TypeError("data must be a pandas DataFrame")

    self.data_missing = data_missing
    self.original = data_original
    self.figsize = (12,12)
    self.fontsize = 10
    self.plot_miss = None
    self.palette = self._set_palette(palette)
    self.custom_color = custom_color

  # ------------------------------------------------------------------------
  def _set_palette(self,palette):
    """
    Sets the color attribute based on the specified palette.

    This method assigns a color tuple to `self.color` based on the provided 
    `palette` name. If `palette` is set to `"customize"`, the method uses 
    the value of `self.custom_color`.

    Parameters
    ----------
    palette : str
        The name of the color palette to use. Must be one of the following: 
        `"gray"`, `"blue"`, `"orange"`, `"green"`, `"red"`, `"customize"`.

    Raises
    ------
    ValueError
        If `palette` is `"customize"` but `self.custom_color` is `None` or 
        not a valid tuple of three numeric values.
        If `palette` is not one of the predefined values.

    Example
    -------
    If `palette` is `"blue"`, `self.color` will be set to `(0, 0, 0.5)`.

    Note
    ----
    - If `palette` is `"customize"`, `self.custom_color` should be a tuple with 
    three elements, each being a float.
    """
    match palette:
      case "gray":
        self.color = (0.25,0.25,0.25)
      case "blue":
        self.color = (0,0,0.5)
      case "orange":
        self.color = (1,0.55,0)
      case "green":
        self.color = (0,0.8,0)
      case "red":
        self.color = (0.8,0,0)
      case "customize":
        if self.custom_color is None:
          raise ValueError("You should set the RGB color for 'customize'.")
        if len(self.custom_color) != 3 or not all(isinstance(c, float) for c in self.custom_color):
            raise ValueError("custom_color must be a tuple with exactly 3 numeric values.")
        self.color = self.custom_color
      case _:
          raise ValueError(f"Unknown palette option: {self.palette}")

    return palette

  # ------------------------------------------------------------------------
  def _histogram(self, col_missing:str, num_bins:int=10):
    """
    Plots a histogram comparing the distribution of a specified column's 
    original data with data containing missing values.

    Parameters
    ----------
    col_missing : str
        The name of the column to plot.
    num_bins : int, optional
        The number of bins to use in the histogram (default is 10).

    Returns
    -------
    fig : matplotlib.figure.Figure
        The matplotlib figure object containing the histogram.
    ax : matplotlib.axes.Axes
        The matplotlib axes object containing the histogram.
    
    """
    data_combined = np.concatenate(
        [self.original[col_missing], self.data_missing[col_missing].dropna(axis=0)]
    )

    bins = np.histogram_bin_edges(data_combined, bins=num_bins)

    fig, ax = plt.subplots(figsize=self.figsize)
    sns.histplot(self.original[col_missing],
                 alpha=0.9,
                 bins=bins,
                 label="Original data",
                 color=self.color)
    sns.histplot(self.data_missing.dropna(axis=0)[col_missing],
                 alpha=0.6,
                 bins=bins,
                 label="Dropping missing values",
                 color="red")
    ax.legend()

    return fig,ax


  # ------------------------------------------------------------------------
  def _boxplot(self,col_missing:str, report:Optional[bool]=False):
    """
    Plots a boxplot comparing the distribution of a specified column's 
    original data with data containing missing values.

    Parameters
    ----------
    col_missing : str
        The name of the column to plot.
    report : bool, optional
        If True, prints the statistical summary of the data (default is False).

    Returns
    -------
    fig : matplotlib.figure.Figure
        The matplotlib figure object containing the boxplot.
    ax : matplotlib.axes.Axes
        The matplotlib axes object containing the boxplot.

    """

    fig, ax = plt.subplots(figsize=self.figsize)
    df_boxplot = pd.DataFrame(columns=["Original data","Dropping missing values"])
    df_boxplot["Original data"] = self.original[col_missing].copy()
    df_boxplot["Dropping missing values"] = self.data_missing.dropna(axis=0)[col_missing].copy()
    df_boxplot.boxplot(color=self.palette)
    ax.set_xlabel(col_missing)

    if report:
      print(df_boxplot.describe())

    return fig,ax


# ------------------------------------------------------------------------
  def _geom_miss_plot(self, cols:Tuple[str, str], threshold:Tuple[float,float,float]=(0.5, 0.8, 0.9)):
      """
      Creates a scatter plot to visualize the relationship between a column 
      with missing values and another column of interest.

      This method generates a scatter plot where the x-axis represents the column 
      with missing values and the y-axis represents another column of interest,
      and vice-versa.

      Parameters
      ----------
      cols : Tuple[str, str]
        The data columns that will be the x-axis and y-axis, respectively.
      threshold : Tuple[float,float,float], optional
        Tuple to control the limits of the jitter of the points.

      Returns
      -------
      fig : matplotlib.figure.Figure
          The matplotlib figure object containing the scatter plot.
      ax : matplotlib.axes.Axes
          The matplotlib axes object containing the scatter plot.

      Reference
      -----
      - https://tmb.njtierney.com/exploring-causes-of-missingness

      """
      
      col_x, col_y = cols
      df_aux = pd.DataFrame(columns=[col_x,col_y])
      df_aux[[col_x, col_y]] = self.data_missing[[col_x, col_y]]
      df_aux['Indicator'] = df_aux[[col_x, col_y]].isna().any(axis=1).map({True:"Missing", False: "Not Missing"})

      min_values = df_aux[[col_x, col_y]].min()
      for col in [col_x, col_y]:
        if df_aux[col].isna().any():
          min_val = min_values[col]
          offset_range = (threshold[0] * min_val, threshold[1] * min_val)
          df_aux[col] = df_aux[col].apply(lambda value: rng.uniform(offset_range[0], offset_range[1]) if pd.isna(value) else value)

      fig,ax = plt.subplots()
      sns.scatterplot(data=df_aux,
                  x=col_x,
                  y=col_y,
                  hue="Indicator",
                  palette={"Missing": "red",
                          "Not Missing": self.color}
                      )   
      plt.xlabel(col_x)
      plt.ylabel(col_y)
      plt.legend()
      plt.axvline(x=threshold[2] * min_values[col_x], color='gray', linestyle='--', linewidth=1.5, alpha=0.5)
      plt.axhline(y=threshold[2] * min_values[col_y], color='gray', linestyle='--', linewidth=1.5, alpha=0.5)

      return fig, ax

  # ------------------------------------------------------------------------
  def visualize_miss(self,
                     visualization_type:str="normal",
                     save:bool=True,
                     path_save_fig:str=None,
                     col_missing:str=None,
                     num_bins:int=10,
                     report:bool=False,
                     threshold=(0.5, 0.8, 0.9),
                     cols:Tuple=None):
    """
    Visualizes the missing data using the specified type of plot and optionally saves the plot.

    Parameters:
    ----------
    visualization_type : str, optional
        The type of visualization to generate. Options are "normal", "bar", "dendrogram",
        "heatmap", "histogram", "boxplot", "scatterplot". Default is "normal".
    save : bool, optional
        Whether to save the plot to a file. Default is True.
    path_save_fig : str, optional
        The path and filename to save the plot. If None, the plot will be saved as 
        <visualization_type>.png. Default is None.
    col_missing : str
        The name of the column to plot.
    num_bins : int, optional
        The number of bins to use in the histogram (default is 10). 
    report : bool, optional
        If True, prints the statistical summary of the data (default is False).
    threshold : Tuple[float,float,float], optional
        Tuple to control the limits of the jitter of the points.
    cols : Tuple[str, str]
      The data columns that will be the x-axis and y-axis, respectively.

    Returns:
    -------
    plot_miss : plt.Axes
        The plot generated by the visualization method.

    References:
    ----------
    missingno documentation: https://github.com/ResidentMario/missingno
    """
    if path_save_fig is None:
      path_save = f"{visualization_type}.png"
    else:
      path_save = path_save_fig

    match visualization_type:
      case "normal":
        self.plot_miss = msno.matrix(self.data_missing,
                    figsize=self.figsize,
                    fontsize=self.fontsize,
                    color = self.color)
        if save:
          plt.savefig(path_save)

      case "bar":
        self.plot_miss = msno.bar(self.data_missing,
                figsize=self.figsize,
                fontsize=self.fontsize,
                color = self.color)
        if save:
          plt.savefig(path_save)

      case "dendrogram":
        self.plot_miss = msno.dendrogram(self.data_missing,
                       figsize=self.figsize,
                       fontsize=self.fontsize)

        if save:
          plt.savefig(path_save)
      case "heatmap":
        self.plot_miss = msno.heatmap(self.data_missing,
                     figsize=self.figsize,
                     fontsize=self.fontsize)
        if save:
          plt.savefig(path_save)

      case "histogram":
        self.plot_miss,ax = self._histogram(col_missing,
                                            num_bins)
        if save:
          self.plot_miss.savefig(path_save)

      case "boxplot":
        self.plot_miss,ax = self._boxplot(col_missing,
                                       report)
        if save:
          self.plot_miss.savefig(path_save)

      case "scatterplot":
        self.plot_miss,ax = self._geom_miss_plot(cols,
                                                 threshold)
        if save:
          self.plot_miss.savefig(path_save)


# -*- coding: utf-8 -*-

# =============================================================================
# Aeronautics Institute of Technologies (ITA) - Brazil
# University of Coimbra (UC) - Portugal
# Arthur Dantas Mangussi - mangussiarthur@gmail.com
# =============================================================================

import numpy as np
import pandas as pd
from sklearn.feature_selection import mutual_info_classif

# ==========================================================================
class MathCalcs:
    # ------------------------------------------------------------------------
    @staticmethod
    def _find_correlation(
        X: pd.DataFrame, y: np.array, colunm: str, flag:bool=False
    ):
        """
        Calculates the correlation matrix of the dataset and finds the feature that is most correlated with the given column.

        Parameters:
        - coluna: str
            The column name to find the most correlated feature with.
        - flag: bool, optional
            If True, find the new most correlated feature excluding the given column and the target feature. Default is False.

        Returns:
        - str
            The name of the most correlated feature.
        """
        (
            matriz_correlacao,
            matriz_correlacao_X,
        ) = MathCalcs._calculate_correlation(X, y)

        id_feature_mais_correlacionada = (
            matriz_correlacao[colunm].drop(colunm).abs().idxmax()
        )

        if flag:
            id_feature_mais_correlacionada = (
                matriz_correlacao_X[colunm].drop(colunm).abs().idxmax()
            )

        return id_feature_mais_correlacionada

    # ------------------------------------------------------------------------
    @staticmethod
    def _nmi(X: pd.DataFrame, y: np.array):
        """
        Calculates the normalized mutual information (NMI) between each feature in a dataset and the target variable.

        Args:
            dataset (pd.DataFrame): The dataset for which to calculate the NMI. It should contain the target variable named "target" and other features.

        Returns:
            List[tuple]: A list of tuples, where each tuple contains the NMI value and the name of the corresponding feature. The list is sorted in descending order of NMI values.
        """
        dataset = X.copy()
        X = X.loc[:, :].values
        dataset['target'] = y

        mi = mutual_info_classif(X, y)

        caracteristicas_ordenadas = sorted(
            zip(map(lambda x: round(x, 4), mi / sum(mi)), dataset.columns),
            reverse=True,
        )

        return [i for i in caracteristicas_ordenadas]

    # ------------------------------------------------------------------------
    @staticmethod
    def _calculate_correlation(X: pd.DataFrame, y: np.array):
        """
        Updates the correlation matrices used in the MissingDataGenerator class.

        This method recalculates the correlation matrix of the original dataset (self.X) and the complete dataset (self.complete_dataset).

        Inputs:
        - None

        Flow:
        1. Calculate the correlation matrix of the original dataset (self.X) and assign it to self.matriz_correlacao_X.
        2. Calculate the correlation matrix of the complete dataset (self.complete_dataset) and assign it to self.matriz_correlacao.

        Outputs:
        - None
        """
        complete_dataset = X.copy()
        complete_dataset['target'] = y

        matriz_correlacao = complete_dataset.corr()

        matriz_correlacao_X = X.corr()

        return matriz_correlacao, matriz_correlacao_X

    # ------------------------------------------------------------------------
    @staticmethod
    def define_groups(dataset, x_obs):
        g1_index = np.random.choice(
            dataset[x_obs].index, round(len(dataset) / 2), replace=False
        )
        g2_index = np.array([i for i in dataset[x_obs].index if i not in g1_index])

        g1 = dataset.loc[g1_index, x_obs]
        g2 = dataset.loc[g2_index, x_obs]

        return g1, g2



# -*- coding: utf-8 -*-

# =============================================================================
# Aeronautics Institute of Technologies (ITA) - Brazil
# University of Coimbra (UC) - Portugal
# Arthur Dantas Mangussi - mangussiarthur@gmail.com
# =============================================================================

import warnings

import numpy as np
import pandas as pd
from scipy.stats import mannwhitneyu
from mdatagen.utils.feature_choice import FeatureChoice


# ==========================================================================
class mMNAR:
    """
    A class to generate missing values in a dataset based on the Missing Not At Random (MNAR) mechanism for multiple features simultaneously.

    Args:
        X (pd.DataFrame): The dataset to receive the missing data.
        y (np.array): The label values from dataset
        missing_rate (int, optional): The rate of missing data to be generated. Default is 10.

    Keyword Args:
        n_xmiss (int, optional): The number of features in the dataset that will receive missing values. Default is the number of features in dataset.
        threshold (float, optional): The threshold to select the locations in feature (xmiss) to receive missing values where 0 indicates de lowest and 1 highest values. Default is 0
        missTarget (bool, optional): A flag to generate missing into the target.

    Example Usage:
    ```python
    # Create an instance of the MNAR class
    generator = MNAR(X, y)

    # Generate missing values using the random strategy
    data_md = generator.random()
    ```
    """

    def __init__(self, X: pd.DataFrame, y: np.array, **kwargs):
        if not isinstance(X, pd.DataFrame):
            raise TypeError('Dataset must be a Pandas Dataframe')
        if not isinstance(y, np.ndarray):
            raise TypeError('y must be a numpy array')
        self.X = X
        self.y = y
        self.dataset = self.X.copy()
        self.missTarget = kwargs.get('missTarget', False)

        if self.missTarget:
            self.dataset['target'] = y

        self.n = self.dataset.shape[0]
        self.p = self.dataset.shape[1]

        self.n_xmiss = kwargs.get('n_xmiss', self.p)
        self.threshold = kwargs.get('threshold', 0)
               

    def random(self, missing_rate: int = 10, deterministic:bool = False):
        """
        Function to randomly choose the feature (x_miss) in dataset for generate missing
        data. The miss locations on x_miss is the lower values based on unobserved feature
        or feature x_miss itself.

        Args:
            missing_rate (int, optional): The rate of missing data to be generated. Default is 10.
            deterministc (bool, optinal): A flag that determine if x_miss will have miss 
            locations based on itself or an unobserved feature. Default is False
            (i.e., an unobserved feature).

        Returns:
            dataset (DataFrame): The dataset with missing values generated under
            the MNAR mechanism.

        Reference:
        [1] Santos, M. S., R. C. Pereira, A. F. Costa, J. P. Soares, J. Santos, and
        P. H. Abreu. 2019. Generating Synthetic Missing Data: A Review by Missing Mechanism.
        IEEE Access 7: 11651–67.
        """

        if missing_rate >= 100:
            raise ValueError(
                'Missing Rate is too high, the whole feature will be deleted!'
            )
        elif missing_rate < 0:
            raise ValueError('Missing rate must be between [0,100]')

        mr = missing_rate / 100
        cont = 0
        N = round((self.n * self.p * mr) / self.n_xmiss)
        xmiss_multiva = []

        if self.n_xmiss > self.p:
            raise ValueError(
                'n_xmiss must be lower than number of feature in dataset'
            )
        elif N > self.n:
            raise ValueError(
                f'FEATURES will be all NaN ({N}>{self.n}), you should increase the number of features that will receive missing data'
            )

        options_xmiss = list(self.dataset.columns.copy())

        while cont < self.n_xmiss:
            # Random choice for determining feature
            x_miss = np.random.choice(options_xmiss)

            if x_miss not in xmiss_multiva:
                x_f = self.dataset.loc[:, x_miss]

                if deterministic:
                    # Observed feature
                    ordered_id = x_f.sort_values()
                    pos_xmiss = FeatureChoice.miss_locations(
                    ordered_id, self.threshold, N
                )

                else:
                    # Unobserved random feature
                    ordered_id = np.lexsort((np.random.random(x_f.size), x_f))
                    pos_xmiss = FeatureChoice.miss_locations(
                    ordered_id, self.threshold, N
                )

                self.dataset.loc[pos_xmiss, x_miss] = np.nan

                xmiss_multiva.append(x_miss)
                options_xmiss.remove(x_miss)
                cont += 1

        if not self.missTarget:
            self.dataset['target'] = self.y
        return self.dataset

    def correlated(self, missing_rate: int = 10, deterministic:bool = False):
        """
        Function to generate missing data in dataset based on correlated pair.
        The feature (x_miss) most correlated with the class for each pair will
        receive the missing data based on lower values of an unobserved feature
        or feature x_miss itself.

        Args:
            missing_rate (int, optional): The rate of missing data to be generated. Default is 10.
            deterministc (bool, optinal): A flag that determine if x_miss will have miss 
            locations based on itself or an unobserved feature. Default is False
            (i.e., an unobserved feature).

        Returns:
            dataset (DataFrame): The dataset with missing values generated under
            the MNAR mechanism.

        Reference:
        [1] Santos, M. S., R. C. Pereira, A. F. Costa, J. P. Soares, J. Santos, and
        P. H. Abreu. 2019. Generating Synthetic Missing Data: A Review by Missing Mechanism.
        IEEE Access 7: 11651–67.
        """

        if missing_rate >= 50:
            raise ValueError(
                'Features will be all NaN, you should decrease the missing rate'
            )

        # Creation of pairs/triples
        pairs,correlation_matrix = FeatureChoice._make_pairs(self.X, self.y, self.missTarget)
        mr = missing_rate / 100

        for pair in pairs:
            if len(pair) % 2 == 0:
                cutK = 2 * mr
                # Find the feature most correlated with the target
                x_miss = FeatureChoice._find_most_correlated_feature_even(
                    correlation_matrix, pair,
                )

            else:
                cutK = 1.5 * mr
                # Find the feature most correlated with the target
                x_miss = FeatureChoice._find_most_correlated_feature_odd(
                    correlation_matrix, pair
                )

            N = round(len(self.dataset) * cutK)

            x_f = self.dataset.loc[:, x_miss]

            if deterministic:
                # Observed feature
                ordered_id = x_f.sort_values()
                pos_xmiss = FeatureChoice.miss_locations(
                ordered_id, self.threshold, N
            )

            else:
                # Unobserved random feature
                ordered_id = np.lexsort((np.random.random(x_f.size), x_f))
                pos_xmiss = FeatureChoice.miss_locations(
                ordered_id, self.threshold, N
            )

            self.dataset.loc[pos_xmiss, x_miss] = np.nan

        if not self.missTarget:
            self.dataset['target'] = self.y
        return self.dataset

    def median(self, missing_rate: int = 10, deterministic:bool = False):
        """
        Function to generate missing data in all dataset based on median from
        each feature. The miss locations are chosen by lower values from a unobserved feature
        or feature x_miss itself.

        Args:
            missing_rate (int, optional): The rate of missing data to be generated. Default is 10.
            deterministc (bool, optinal): A flag that determine if x_miss will have miss 
            locations based on itself or an unobserved feature. Default is False
            (i.e., an unobserved feature).

        Returns:
            dataset (DataFrame): The dataset with missing values generated under
            the MNAR mechanism.

        Reference:
        [1] Santos, M. S., R. C. Pereira, A. F. Costa, J. P. Soares, J. Santos, and
        P. H. Abreu. 2019. Generating Synthetic Missing Data: A Review by Missing Mechanism.
        IEEE Access 7: 11651–67.
        """

        if missing_rate >= 50:
            raise ValueError(
                'Features will be all NaN, you should decrease the missing rate'
            )

        mr = missing_rate / 100

        # For each column in dataset
        for col in self.dataset.columns:
            N = round(len(self.dataset) * mr)

            if len(self.dataset[col].unique()) == 2:  # Binary feature
                g1_index = np.random.choice(
                    self.dataset[col].index,
                    round(len(self.dataset) / 2),
                    replace=False,
                )
                g2_index = np.array([i for i in self.dataset[col].index if i not in g1_index])

            else:  # Continuos or ordinal feature
                median_xobs = self.dataset[col].median()

                g1 = self.dataset[col][self.dataset[col] <= median_xobs]
                g2 = self.dataset[col][self.dataset[col] > median_xobs]

                # If median do not divide in equal-size groups
                if len(g1) != len(g2):  
                    g1_index = np.random.choice(
                        self.dataset[col].index,
                        round(len(self.dataset) / 2),
                        replace=False,
                    )
                    g2_index = np.array([i for i in self.dataset[col].index if i not in g1_index])
                else:
                    g1_index = g1.index
                    g2_index = g2.index

            choice = np.random.choice([0, 1])
            if choice == 0:
                x_f = self.dataset.loc[g1_index, col]
            else:
                x_f = self.dataset.loc[g2_index, col]

            if deterministic:
                # Observed feature
                ordered_id = x_f.sort_values()
                pos_xmiss = FeatureChoice.miss_locations(
                ordered_id, self.threshold, N
            )

            else:
                # Unobserved random feature
                ordered_id = np.lexsort((np.random.random(x_f.size), x_f))
                pos_xmiss = FeatureChoice.miss_locations(
                ordered_id, self.threshold, N
            )

            self.dataset.loc[pos_xmiss, col] = np.nan

        if not self.missTarget:
            self.dataset['target'] = self.y
        return self.dataset

    def MBOUV(
        self, missing_rate: int = 10, depend_on_external=None, ascending=True
    ):
        """
        Function to generate missing data based on Missigness Based on Own and Unobserved Values (MBOUV).

        Args:
            missing_rate (int, optional): The rate of missing data to be generated. Default is 10.

        Returns:
            dataset (DataFrame): The dataset with missing values generated under
            the MNAR mechanism.

        Reference:
        [2] Pereira, R. C., Abreu, P. H., Rodrigues, P. P., Figueiredo, M. A. T., (2024). 
        Imputation of data Missing Not at Random: Artificial generation and benchmark analysis. 
        Expert Systems with Applications, 249(B), 123654.


        """

        if depend_on_external is None:
            depend_on_external = []

        if missing_rate > 90:
            raise ValueError(
                'Maximum missing rate per feature must be clipped at 90%'
            )
        elif missing_rate < 0:
            raise ValueError('Missing rate must be between [0,100]')

        mr = missing_rate / 100
        N = round(mr * self.n * self.p)

        # With moderate missing rates (up to 60%), it never happens.
        max_mvs_feat = round(min(mr + (mr / 2.0), 0.9) * self.n)

        # Randomly distribute missing values by the features.
        num_mvs_per_feat = np.zeros(self.p)
        while N > 0:
            for i in range(self.p):
                if num_mvs_per_feat[i] < max_mvs_feat:
                    num_mvs_it = np.random.randint(
                        0, min(max_mvs_feat - num_mvs_per_feat[i] + 1, N + 1)
                    )
                    N -= num_mvs_it
                    num_mvs_per_feat[i] += num_mvs_it
                    if N == 0:
                        break

        np.random.shuffle(num_mvs_per_feat)
        hidden_f = np.random.normal(size=self.n)
        tie_breaker = np.random.random(hidden_f.size)

        # Amputate the values.
        for i, col in enumerate(self.dataset.columns.values):
            num_mv = round(num_mvs_per_feat[i])
            num_mv = num_mv if num_mv > 0 else 0

            if col in depend_on_external:
                start_n = end_n = int(num_mv / 2)
                if num_mv % 2 == 1:
                    end_n += 1

                indices_start = np.lexsort((tie_breaker, hidden_f))[:start_n]
                indices_end = np.lexsort((tie_breaker, -hidden_f))[:end_n]

                self.dataset.loc[indices_start, col] = np.nan
                self.dataset.loc[indices_end, col] = np.nan
            else:
                ordered_indices = FeatureChoice.get_ordered_indices(
                    col, self.dataset, ascending
                )
                self.dataset.loc[ordered_indices[:num_mv], col] = np.nan

        if not self.missTarget:
            self.dataset['target'] = self.y
        return self.dataset

    def MBOV_randomness(
        self,
        missing_rate: int = 10,
        randomness: float = 0,
        columns: list = None
    ):
        """
        Function to generate missing data based on Missigness Based on Own Values (MBOV) using
        a randomess to choose miss locations in each feature.

        Args:
            missing_rate (int, optional): The rate of missing data to be generated. Default is 10.
            randomness (float, optional): The randomness rate for choose miss locations. Default is 0 that represents lower values
            columns (list): A list of strings containing columns names.

        Returns:
            dataset (DataFrame): The dataset with missing values generated under
            the MNAR mechanism.

        Reference:
        [2] Pereira, R. C., Abreu, P. H., Rodrigues, P. P., Figueiredo, M. A. T., (2024). 
        Imputation of data Missing Not at Random: Artificial generation and benchmark analysis. 
        Expert Systems with Applications, 249(B), 123654.

        """
        if not (0 <= randomness <= 0.5):
            raise ValueError('randomness must be in range [0,0.5]')

        if columns is None:
            raise TypeError(
                'You must inform columns from dataset to generate missing data'
            )

        if missing_rate >= 100:
            raise ValueError(
                'Missing Rate is too high, the whole feature will be deleted!'
            )
        elif missing_rate < 0:
            raise ValueError('Missing rate must be between [0,100]')

        mr = missing_rate / 100
        N = round(len(self.dataset) * mr)
        aux_index = list(self.dataset.index)

        for col in columns:
            pos_xmis_deterministic = (
                self.dataset[col]
                .sort_values()[: round((1 - randomness) * N)]
                .index
            )

            # Eliminate the lowest candidate from index
            np.delete(aux_index, pos_xmis_deterministic)

            pos_xmiss_randomness = np.random.choice(
                aux_index, round(randomness * N), replace=False
            )

            pos_xmis = np.concatenate(
                [pos_xmis_deterministic, pos_xmiss_randomness]
            )

            self.dataset.loc[pos_xmis, col] = np.nan

        if not self.missTarget:
            self.dataset['target'] = self.y
        return self.dataset

    def MBOV_median(self, missing_rate: int = 10, columns: list = None):
        """
        Function to generate missing data based on Missigness Based on Own Values (MBOV) using
        a median to choose miss locations in each feature.

        Args:
            missing_rate (int, optional): The rate of missing data to be generated. Default is 10.
            columns (list): A list of strings containing columns names.

        Returns:
            dataset (DataFrame): The dataset with missing values generated under
            the MNAR mechanism.

        Reference:
        [2] Pereira, R. C., Abreu, P. H., Rodrigues, P. P., Figueiredo, M. A. T., (2024). 
        Imputation of data Missing Not at Random: Artificial generation and benchmark analysis. 
        Expert Systems with Applications, 249(B), 123654.

        """

        if missing_rate >= 100:
            raise ValueError(
                'Missing Rate is too high, the whole feature will be deleted!'
            )
        elif missing_rate < 0:
            raise ValueError('Missing rate must be between [0,100]')

        if columns is None:
            raise TypeError(
                'You must inform columns from dataset to generate missing data'
            )

        mr = missing_rate / 100
        N = round(len(self.dataset) * mr)

        for col in columns:
            if self.dataset[col].dtype == 'object':
                raise TypeError(
                    'Only continuos or ordinal feature are accepted'
                )

            median_att = self.dataset[col].median()

            pos_xmis = np.argsort(np.abs(self.dataset[col] - median_att))[:N]
            self.dataset.loc[pos_xmis, col] = np.nan

        if not self.missTarget:
            self.dataset['target'] = self.y
        return self.dataset

    def MBIR(
        self,
        missing_rate: int = 10,
        columns: list = None,
        statistical_method: str = 'Mann-Whitney',
    ):
        """
        Function to generate missing data based on Missingness Based on Intra-Relation (MBIR).

        Args:
            missing_rate (int, optional): The rate of missing data to be generated. Default is 10.
            columns (list): A list of strings containing columns names.
            statistical_method (str, optional): A string to inform statistical method. 
            The options are ["Mann-Whitney", "Bayesian"]. Default is Mann-Whitney

        Returns:
            dataset (DataFrame): The dataset with missing values generated under
            the MNAR mechanism.

        Reference:
        [2] Pereira, R. C., Abreu, P. H., Rodrigues, P. P., Figueiredo, M. A. T., (2024). 
        Imputation of data Missing Not at Random: Artificial generation and benchmark analysis. 
        Expert Systems with Applications, 249(B), 123654.

        """
        if missing_rate >= 100:
            raise ValueError(
                'Missing Rate is too high, the whole feature will be deleted!'
            )
        elif missing_rate < 0:
            raise ValueError('Missing rate must be between [0,100]')

        if columns is None:
            raise TypeError(
                'You must inform columns from dataset to generate missing data'
            )

        elif set(columns) == set(self.dataset):
            warnings.warn('All dataset will be drop', UserWarning)

        mr = missing_rate / 100
        N = round(len(self.dataset) * mr)

        df_columns = self.dataset.columns.tolist()

        for x_miss in columns:
            most_significant_diff = {}
            for x_obs in df_columns:
                if self.dataset[x_obs].dtype != 'object':

                    instances = self.dataset.copy()

                    if x_obs != x_miss:
                        # Find the instance with lower MR% values of x_obs
                        lower = instances[x_obs].sort_values()[:N].index

                        # Set the f_miss to be missing
                        instances.loc[lower, x_miss] = np.nan

                        # Create the missing indicator
                        auxiliary_ind = np.where(
                            instances[x_miss].isna(), 1, 0
                        )

                        # Statistical tests between values of f_obs and f_ind
                        match statistical_method:
                            case 'Mann-Whitney':
                                statistic, p_value = mannwhitneyu(
                                    instances[x_obs], auxiliary_ind
                                )

                        if p_value < 0.05:
                            # There is evidence of a significant difference.
                            most_significant_diff[x_obs] = p_value

            most_feature = min(
                most_significant_diff, key=most_significant_diff.get
            )
            pos_xmis = self.dataset[most_feature].sort_values()[:N].index
            self.dataset.loc[pos_xmis, x_miss] = np.nan
            df_columns.remove(most_feature)
            self.dataset = self.dataset.drop(columns=most_feature)

        if not self.missTarget:
            self.dataset['target'] = self.y
        return self.dataset

if __name__ == "__main__":
    import numpy as np 
    import pmlb
    from time import perf_counter

    def split_data(data):
        df = data.copy()
        X = df.drop(columns=["target"])
        y = data["target"]

        return X,np.array(y)

    kddcup = pmlb.fetch_data('kddcup')
    X_, y_ = split_data(kddcup)

    time_init = perf_counter()
    generator = mMNAR(X=X_, y=y_, missTarget=True)
    gen_md = generator.correlated(missing_rate=45)


# -*- coding: utf-8 -*-

# =============================================================================
# Aeronautics Institute of Technologies (ITA) - Brazil
# University of Coimbra (UC) - Portugal
# Arthur Dantas Mangussi - mangussiarthur@gmail.com
# =============================================================================

import numpy as np
import pandas as pd
from mdatagen.utils.math_calcs import MathCalcs

# ==========================================================================
class FeatureChoice:
    @staticmethod
    def _define_xmiss(
        X: pd.DataFrame, y: np.array, x_miss: str = None, x_obs:str = None, flag:bool=True
    ):
        if not x_miss:
            x_miss = MathCalcs._find_correlation(X, y, 'target')

        if not x_obs:
            if flag:
                x_obs = MathCalcs._find_correlation(X, y, x_miss)

                if x_obs == 'target':
                    x_obs = MathCalcs._find_correlation(X, y, x_miss, flag=True)

        return x_miss, x_obs

    # ------------------------------------------------------------------------
    @staticmethod
    def get_ordered_indices(col, dataset, ascending):
        x_f = dataset.loc[:, col].values
        tie_breaker = np.random.random(x_f.size)
        if ascending:
            return np.lexsort((tie_breaker, x_f))
        else:
            return np.lexsort((tie_breaker, x_f))[::-1]

    # ------------------------------------------------------------------------
    @staticmethod
    def _delete(item: str, list_items: np.array) -> np.array:
        """
        Remove the specified item from the list of items.

        Args:
            item (str): The item to be deleted from the list.
            list_items (np.array): The list of items from which the specified item will be removed.

        Returns:
            np.array: The updated list of items with the specified item removed.
        """
        mask = list_items != item
        return list_items[mask]

    # ------------------------------------------------------------------------
    @staticmethod
    def _add_var_remaining(pairs: list, var: str, col: str) -> list:
        """
        Add the column to the pair if the variable is present in the pair.

        Args:
            pairs (list): A list of tuples representing pairs of variables.
            var (str): The variable to check for in the pairs.
            col (str): The column to add to the pairs.

        Returns:
            list: The updated list of pairs with the col added to the tuples where var is present.
        """
        for i, tpl in enumerate(pairs):
            if var in tpl:
                pairs[i] = tpl + (col,)

        return pairs

    # ------------------------------------------------------------------------
    @staticmethod
    def _make_pairs(X: pd.DataFrame, y: np.ndarray, missTarget:bool=False)->list[tuple]:
        df = X.copy()
        df['target'] = y
        
        # If correlation is NaN, we set to 0
        correlation_matrix = df.corr().fillna(0)       
        
        if missTarget:
            # Flatten the correlation matrix and exclude self-correlations
                correlations = (
                    correlation_matrix.where(correlation_matrix != 1.0)
                    .stack()
                    .reset_index()
                )
        else:
            correlations = (
                correlation_matrix.drop(columns="target", index="target").where(correlation_matrix != 1.0)
                .stack()
                .reset_index()
            )
            
        correlations.columns = ["Feature 1", "Feature 2", "Correlation"]

        # Sort by absolute correlation in descending order
        correlations = correlations.sort_values(by="Correlation", key=abs, ascending=False).reset_index(drop=True)

        # Initialize sets to track paired features and the final pairs
        paired_features = set()
        pairs = []

        # Track the feature that is left out (if any)
        unpaired_feature = None
        
        # Iterate through the sorted correlations to find the best pairs
        for _, row in correlations.iterrows():
            f1, f2 = row["Feature 1"], row["Feature 2"]
            if f1 not in paired_features and f2 not in paired_features:
                pairs.append((f1, f2))
                paired_features.update([f1, f2])

        # Identify the unpaired feature, if there is any
        all_features = set(correlations['Feature 1']).union(set(correlations['Feature 2']))
        unpaired_feature = list(all_features - paired_features)

        # If there is an unpaired feature, add it to the pair with the highest correlation
        if unpaired_feature:
            # The remaining highest correlation pair is the first in the sorted list
            remaining_pair = pairs[0]  # Pair with the highest correlation (already sorted)
            # Add the unpaired feature to this pair
            pairs[0] = (remaining_pair[0], remaining_pair[1], unpaired_feature[0])

        return pairs, correlation_matrix
            
    # ------------------------------------------------------------------------
    @staticmethod
    def _find_most_correlated_feature_even(
        correlation_matrix:pd.DataFrame, pair_features:list[tuple]
    ):
        
        corr_0 = abs(correlation_matrix['target'][pair_features[0]])
        corr_1 = abs(correlation_matrix['target'][pair_features[1]])

        return pair_features[0] if corr_0 > corr_1 else pair_features[1]

    # ------------------------------------------------------------------------
    @staticmethod
    def _find_most_correlated_feature_odd(
        correlation_matrix:pd.DataFrame, pair_features:list[tuple]
    ):
        
        corr_0 = abs(correlation_matrix['target'][pair_features[0]])
        corr_1 = abs(correlation_matrix['target'][pair_features[1]])
        corr_2 = abs(correlation_matrix['target'][pair_features[2]])

        correlations = [
            (pair_features[0], corr_0),
            (pair_features[1], corr_1),
            (pair_features[2], corr_2),
        ]

        sorted_correlations = sorted(
            correlations, key=lambda x: x[1], reverse=True
        )

        return [t[0] for t in sorted_correlations][:2]

    # ------------------------------------------------------------------------
    @staticmethod
    def miss_locations(ordered_id, threshold, N):

        lowest = ordered_id[: round((1 - threshold) * N)]

        highest = [] if threshold == 0 else ordered_id[round(-threshold * N) :]

        if isinstance(ordered_id, pd.Series):
            if len(lowest) == 0:
                pos_xmiss = np.hstack([lowest, highest.index])
            elif len(highest) == 0:
                pos_xmiss = np.hstack([lowest.index, highest])
            else:
                pos_xmiss = np.hstack([lowest.index, highest.index])

        elif isinstance(ordered_id, np.ndarray):
            pos_xmiss = np.hstack([lowest, highest])
        else:
            raise TypeError('ordered_id must be pd.Series or np.array')

        return pos_xmiss



# -*- coding: utf-8 -*

#  =============================================================================
# Aeronautics Institute of Technologies (ITA) - Brazil
# University of Coimbra (UC) - Portugal
# Arthur Dantas Mangussi - mangussiarthur@gmail.com
# =============================================================================

import pandas as pd
from typing import Optional
from sklearn.metrics import (mean_absolute_error,
                             mean_squared_error,
                             mean_absolute_percentage_error,
                             root_mean_squared_error,
                             r2_score
                            )
from scipy.stats import kstest, pearsonr
import numpy as np

# ==========================================================================
class EvaluateImputation:
    """
    A class to evaluate the performance of imputed data against the 
    original data using various metrics.

    Attributes
    ----------
    original : pd.DataFrame
        The original data before imputation.
    data_imputed : pd.DataFrame
        The data after imputation.
    metric : str, optional. Default is "mean_absolute_error".
        The metric to use for evaluating the imputed data. Supported metrics are:
        - "mean_absolute_error"
        - "mean_squared_error"
        - "root_mean_squared_error"
        - "kolmogorov_smirnov_distance" 
        - "pearson_correlation_coefficient" 
        - "proportion_falsely_classified_entries"
        - "mean_absolute_percentage_error" 
        - "normalize_mean_absolute_error"
        - "percentage_of_correct_prediction"
        - "R2"
    data_missing : pd.DataFrame, optional
        The data with missing values. Currently supports as a parameter
        for "proportion_falsely_classified_entries". 
    
    Example:
    import pandas as pd
    from sklearn.datasets import load_iris

    from mdatagen.univariate.uMCAR import uMCAR

    # Load the data
    iris = load_iris()
    iris_df = pd.DataFrame(data=iris.data, columns=iris.feature_names)

    X = iris_df.copy()   # Features
    y = iris.target    # Label values

    # Create a instance with missing rate equal to 25% in dataset under MCAR mechanism
    generator = uMCAR(X=X, y=y, missing_rate=25, x_miss='petal length (cm)')

    # Generate the missing data under MNAR mechanism
    generate_data = generator.random()

    eval_metric = EvaluateImputation(data_imputed=generate_data.drop("target",axis=1).fillna(0),
                                     data_original=X)
    eval_metric.show()

    """
    def __init__(self, 
                 data_imputed:pd.DataFrame,
                 data_original:pd.DataFrame, 
                 metric:Optional[str] = "mean_absolute_error",
                 data_missing:pd.DataFrame = None
                 ) -> None:
        
        if data_imputed.shape != data_original.shape:
            raise ValueError("Different shapes for the original and imputed data")
        self.original = data_original
        self.data_imputed = data_imputed
        self.metric = metric
        self.data_missing = data_missing
        
    # ------------------------------------------------------------------------
    def _choose_metric(self):
        """
        Selects and executes the appropriate metric function based on the provided metric.

        Returns
        -------
        float
            The result of the selected error metric.

        """

        metrics = {
            "mean_absolute_error": self._mean_absolute_error,
            "mean_squared_error": self._mean_squared_error,
            "root_mean_squared_error": self._root_mean_squared_error,
            "kolmogorov_smirnov_distance": self._kolmogorov_smirnov_distance,
            "pearson_correlation_coefficient": self._pearson_correlation_coefficient,
            "proportion_falsely_classified_entries": self._proportion_falsely_classified_entries,
            "mean_absolute_percentage_error": self._mean_absolute_percentage_error,
            "normalize_mean_absolute_error": self._normalize_mean_absolute_error,
            "percentage_of_correct_prediction": self._percentage_of_correct_prediction,
            "R2": self._r2_Score
        }
        
        self.metric_function = metrics.get(self.metric)

        if self.metric_function:
            error_metric_result = self.metric_function()
            return error_metric_result
        else:
            raise ValueError(f"Metric '{self.metric}' is not supported.")
    
    # ------------------------------------------------------------------------
    def show(self):
        """
        Calculates and returns the selected error metric.

        Returns
        -------
        float
            The result of the selected error metric.
        """
        return self._choose_metric()

    # ------------------------------------------------------------------------
    def _mean_absolute_error(self):
        """
        Calculates and returns the Mean Absolute Error (MAE).
        Values closer to 0 indicate more accurate imputations.

        Returns
        -------
        float
            The Mean Absolute Error (MAE) between the original and imputed data.
        
        Reference:
        [1] Hasan, Md. K., Alam, Md. A., Roy, S., Dutta, A., Jawad, Md. T., and Das, S. 2021. 
        Missing value imputation affects the performance of machine learning: A review and 
        analysis of the literature (2010–2021). Informatics in Medicine Unlocked. 27. 
        10.1016/j.imu.2021.100799. 
        """
        mae = mean_absolute_error(y_true=self.original,
                                  y_pred=self.data_imputed)
        return mae
    
    # ------------------------------------------------------------------------
    def _mean_squared_error(self):
        """
        Calculates and returns the Mean Squared Error (MSE).
        Values closer to 0 indicate more accurate imputations.

        Returns
        -------
        float
            The Mean Squared Error (MSE) between the original and imputed data.
        
        Reference:
        [1] Hasan, Md. K., Alam, Md. A., Roy, S., Dutta, A., Jawad, Md. T., and Das, S. 2021. 
        Missing value imputation affects the performance of machine learning: A review and 
        analysis of the literature (2010–2021). Informatics in Medicine Unlocked. 27. 
        10.1016/j.imu.2021.100799. 
        """
        mse = mean_squared_error(y_true=self.original,
                                  y_pred=self.data_imputed)
        return mse
    
    # ------------------------------------------------------------------------
    def _root_mean_squared_error(self):
        """
        Calculates and returns the Root Mean Squared Error (RMSE).
        Values closer to 0 indicate more accurate imputations.

        Returns
        -------
        float
            The Root Mean Squared Error (RMSE) between the original and imputed data.

        Reference:
        [2] Lin, W. C., and Tsai, C. F. 2020. Missing value imputation: a review and analysis of the 
        literature (2006–2017). Artif Intell Rev 53, 1487–1509. 
        https://doi.org/10.1007/s10462-019-09709-4
        """
        rmse = root_mean_squared_error(y_true=self.original,
                                  y_pred=self.data_imputed,
                                  )
        return rmse
    
    # ------------------------------------------------------------------------
    def _kolmogorov_smirnov_distance(self):
        """
        Calculates and returns the Kolmogorov-Smirnov Distance.

        Returns
        -------
        float
            The Kolmogorov-Smirnov Distance between the original and imputed data.
        
        Reference:
        [3] Nguyen, C. D., Carlin, J. B., and Lee, K.J. 2013. Diagnosing problems with 
        imputation models using the Kolmogorov-Smirnov test: a simulation study. 
        BMC Med Res Methodol 13, 144. 
        https://doi.org/10.1186/1471-2288-13-144
        """
        original = np.array(self.original).flatten()
        imputed = np.array(self.data_imputed).flatten()
        kolmogorov = kstest(rvs=original,
                            cdf=imputed)
        return kolmogorov
    
    # ------------------------------------------------------------------------
    def _pearson_correlation_coefficient(self):
        """
        Calculates and returns the Pearson Correlation Coefficient.

        Returns
        -------
        float
            The Pearson Correlation Coefficient between the original and imputed data.
        
        Reference:
        [4] 
        """
        original = np.array(self.original).flatten()
        imputed = np.array(self.data_imputed).flatten()
        pearson = pearsonr(original,
                           imputed)
        return pearson
    
    # ------------------------------------------------------------------------
    def _proportion_falsely_classified_entries(self):
        """
        Calculates and returns the Proportion of Falsely Classified Entries.

        Returns
        -------
        float
            The Proportion of Falsely Classified Entries between the original and imputed data.

        Reference:
        [5] Guo, C. Y., Yang, Y. C., and Chen, Y. H. 2021. The Optimal Machine Learning-Based Missing Data 
        Imputation for the Cox Proportional Hazard Model. Frontiers in public health, 9, 680054.
        https://doi.org/10.3389/fpubh.2021.680054

        """
        if self.data_missing is None:
            raise ValueError("You have to provide the data with missing values")
        
        original = np.array(self.original).flatten()
        imputed = np.array(self.data_imputed).flatten() 
        not_identical = sum(a != b for a, b in zip(original, imputed))
        pfc = not_identical/sum(np.isnan(np.array(self.data_missing).flatten()))

        return pfc
    
    # ------------------------------------------------------------------------
    def _mean_absolute_percentage_error(self):
        """
        Calculates and returns the Mean Absolute Percentage Error (MAPE).
        Values closer to 0 indicate more accurate imputations.

        Returns
        -------
        float
            The Mean Absolute Percentage Error (MAPE) between the original and imputed data.
        
        Reference:
        [2] Lin, W. C., and Tsai, C. F. 2020. Missing value imputation: a review and analysis of the 
        literature (2006–2017). Artif Intell Rev 53, 1487–1509. 
        https://doi.org/10.1007/s10462-019-09709-4.
        """
        mape = mean_absolute_percentage_error(y_true=self.original,
                                  y_pred=self.data_imputed)
        return mape
    
    # ------------------------------------------------------------------------
    def _normalize_mean_absolute_error(self):
        """
        Calculates and returns the Normalize Mean Absolute Error (NMAE).

        Returns
        -------
        float
            The Normalize Mean Absolute Error (NMAE) between the original and imputed data.
        
        Reference:
        [6] 
        """
        mae = mean_absolute_error(y_true=self.original,
                                  y_pred=self.data_imputed)
        
        max_xj = max(self.original)
        min_xj = min(self.original)
        nmae = mae / (max_xj - min_xj)
        
        return nmae
    
    # ------------------------------------------------------------------------
    def _percentage_of_correct_prediction(self):
        """
        Calculates and returns the Percentage of Correct Predictions (PCP).
        Values closer to 1 indicate more accurate imputations.

        Returns
        -------
        float
            The Percentage of Correct Predictions (PCP) between the original and imputed data.

        Reference:
        [2] Lin, W. C., and Tsai, C. F. 2020. Missing value imputation: a review and analysis of the 
        literature (2006–2017). Artif Intell Rev 53, 1487–1509. 
        https://doi.org/10.1007/s10462-019-09709-4
        """
        original = np.array(self.original).flatten()
        imputed = np.array(self.data_imputed).flatten() 
        correct = sum(a == b for a, b in zip(original, imputed))
        pcp = correct / len(self.original.shape[0])

        return round(100 * pcp, 2)
    
    # ------------------------------------------------------------------------
    def _r2_Score(self):
        """
        Calculates and returns the R2 score (coefficient of determination).

        Returns
        -------
        float
            The R2 (coefficient of determination) between the original and imputed data.
        
        Reference:
        [1] Hasan, Md. K., Alam, Md. A., Roy, S., Dutta, A., Jawad, Md. T., and Das, S. 2021. 
        Missing value imputation affects the performance of machine learning: A review and 
        analysis of the literature (2010–2021). Informatics in Medicine Unlocked. 27. 
        10.1016/j.imu.2021.100799. 
        """
        r2 = r2_score(y_true=self.original,
                      y_pred=self.data_imputed)

        return r2

    
    


# -*- coding: utf-8 -*

# =============================================================================
# Aeronautics Institute of Technologies (ITA) - Brazil
# University of Coimbra (UC) - Portugal
# Arthur Dantas Mangussi - mangussiarthur@gmail.com
# =============================================================================

import warnings

import numpy as np
import pandas as pd


# ==========================================================================
class mMCAR:
    """
    A class to generate missing data in a dataset based on the Missing Completely At Random (MCAR) mechanism for multiple features simultaneously.

    Args:
        X (pd.DataFrame): The dataset to receive the missing data.
        y (np.array): The label values from dataset
        missing_rate (int, optional): The rate of missing data to be generated. Default is 10.
        missTarget (bool, optional): A flag to generate missing into the target.
        seed (int, optional): The seed for the random number generator.

    Example Usage:
    ```python
    # Create an instance of the MCAR class
    generator = MCAR(X, y, missing_rate=20)

    # Generate missing values using the random strategy
    data_md = generator.random()
    ```
    """

    def __init__(self, X: pd.DataFrame, y: np.array, missing_rate: int = 10, missTarget:bool=False, seed:int=None):
        if not isinstance(X, pd.DataFrame):
            raise TypeError('Dataset must be a Pandas Dataframe')
        if not isinstance(y, np.ndarray):
            raise TypeError('y must be a numpy array')

        if missing_rate >= 100:
            raise ValueError(
                'Missing Rate is too high, the whole dataset will be deleted!'
            )
        elif missing_rate < 0:
            raise ValueError('Missing rate must be between [0,100]')

        self.X = X
        self.y = y
        self.dataset = self.X.copy()
        self.missing_rate = missing_rate
        self.missTarget = missTarget

        self.seed = seed
        if seed is not None:
            np.random.seed(seed)

        if missTarget:
            self.dataset['target'] = y

    def random(self) -> pd.DataFrame:
        """
        Function to randomly generate missing data in all dataset.

        Returns:
            dataset (DataFrame): The dataset with missing values generated under
            the MCAR mechanism.

        Reference:
        [1] Santos, M. S., R. C. Pereira, A. F. Costa, J. P. Soares, J. Santos, and
        P. H. Abreu. 2019. Generating Synthetic Missing Data: A Review by Missing Mechanism.
        IEEE Access 7: 11651–67.

        """
        original_shape = self.dataset.shape
        mr = self.missing_rate / 100
        n = self.dataset.shape[0]
        p = self.dataset.shape[1]
        N = round(n * p * mr)

        array_values = self.dataset.values
        pos_miss = np.random.choice(
            self.dataset.shape[0] * self.dataset.shape[1], N, replace=False
        )
        array_values = array_values.flatten()
        array_values[pos_miss] = np.nan
        array_values = array_values.reshape(original_shape)

        return pd.DataFrame(array_values, columns=self.dataset.columns)

    def binomial(self, columns: list = None):
        """
        Function to generate missing data in columns by Bernoulli distribution
        for each attribute informed.

        Args:
            columns (list): A list of strings containing columns names.

        Returns:
            dataset (DataFrame): The dataset with missing values generated under
            the MCAR mechanism.

        Reference:
        [1] Santos, M. S., R. C. Pereira, A. F. Costa, J. P. Soares, J. Santos, and
        P. H. Abreu. 2019. Generating Synthetic Missing Data: A Review by Missing Mechanism.
        IEEE Access 7: 11651–67.

        """

        warnings.warn(
            'Binomial sometimes does not generate the input missing rate in dataset'
        )

        if columns is None:
            raise TypeError(
                'You must inform columns from dataset to generate missing data'
            )

        for x_miss in columns:
            pos_xmiss = np.random.binomial(
                n=1, p=self.missing_rate / 100, size=self.dataset.shape[0]
            ).astype(bool)
            self.dataset.loc[pos_xmiss, x_miss] = np.nan
        
        if not self.missTarget:
            self.dataset['target'] = self.y
        return self.dataset



# -*- coding: utf-8 -*

# =============================================================================
# Aeronautics Institute of Technologies (ITA) - Brazil
# University of Coimbra (UC) - Portugal
# Arthur Dantas Mangussi - mangussiarthur@gmail.com
# =============================================================================

import random

import numpy as np
import pandas as pd
from mdatagen.utils.feature_choice import FeatureChoice

# ==========================================================================
class uMAR:
    """
    A class to generate missing values in a dataset based on the Missing At Random (MAR) univariate mechanism.

    Args:
        X (pd.DataFrame): The dataset to receive the missing data.
        y (np.array): The label values from dataset
        missing_rate (int, optional): The rate of missing data to be generated. Default is 10.
        x_miss (string): The name of feature to insert the missing data. If not informed, x_miss will be the feature most correlated with target
        x_obs (string): The name of observed feature. If not informed, x_obs will be the feature most correlated with x_miss.
    Example Usage:
    ```python
    # Create an instance of the MAR class
    generator = MAR(X, y, missing_rate=20, x_miss='feature1')

    # Generate missing values using the lowest strategy
    data_md = generator.lowest()
    ```
    """

    # ------------------------------------------------------------------------
    def __init__(
        self,
        X: pd.DataFrame,
        y: np.array,
        missing_rate: int = 10,
        x_miss: str = None,
        x_obs:str = None
    ):
        if not isinstance(X, pd.DataFrame):
            raise TypeError('Dataset must be a Pandas Dataframe')
        if not isinstance(y, np.ndarray):
            raise TypeError('y must be a numpy array')

        self.missing_rate = missing_rate

        if self.missing_rate >= 100:
            raise ValueError(
                'Missing Rate is too high, the whole feature will be deleted!'
            )
        elif self.missing_rate < 0:
            raise ValueError('Missing rate must be between [0,100]')

        self.X = X
        self.y = y
        self.dataset = self.X.copy()
        self.dataset['target'] = y

        self.x_miss, self.x_obs = FeatureChoice._define_xmiss(
            self.X, self.y, x_miss, x_obs
        )
        self.N = round(self.missing_rate * self.dataset.shape[0] / 100)

    # ------------------------------------------------------------------------
    def lowest(self):
        """Function to generate missing values in the feature (x_miss) using
        the lowest values from an observed feature.

        Returns:
            dataset (DataFrame): The dataset with missing values generated under
            the MAR mechanism.

        Reference:
        [1] Santos, M. S., R. C. Pereira, A. F. Costa, J. P. Soares, J. Santos, and
        P. H. Abreu. 2019. Generating Synthetic Missing Data: A Review by Missing Mechanism.
        IEEE Access 7: 11651–67.

        """
        pos_xmis = self.dataset[self.x_obs].sort_values()[: self.N].index
        self.dataset.loc[pos_xmis, self.x_miss] = np.nan
        return self.dataset

    # ------------------------------------------------------------------------
    def rank(self):
        """
        Function to generate missing values in the feature (x_miss) using
        a rank from an observed feature.

        Returns:
            dataset (DataFrame): The dataset with missing values generated under
            the MAR mechanism.

        Reference:
        [1] Santos, M. S., R. C. Pereira, A. F. Costa, J. P. Soares, J. Santos, and
        P. H. Abreu. 2019. Generating Synthetic Missing Data: A Review by Missing Mechanism.
        IEEE Access 7: 11651–67.

        """
        ranks = self.dataset[self.x_obs].rank(
            method='average', na_option='bottom'
        )
        random_numbers = np.random.random(self.dataset.shape[0])
        lista_index = set()
        i, max_iter = 0, 0
        while i < self.N:
            if max_iter == 50:
                random_numbers = np.random.random(self.dataset.shape[0])

            lin = random.randint(0, self.dataset.shape[0] - 1)
            if (
                random_numbers[lin] <= ranks[lin] / (ranks.max() + 1)
                and lin not in lista_index
            ):
                lista_index.add(lin)
                max_iter = 0
                i += 1
            else:
                max_iter += 1
        pos_xmis = list(lista_index)
        self.dataset.loc[pos_xmis, self.x_miss] = np.nan
        return self.dataset

    # ------------------------------------------------------------------------
    def median(self):
        """
        Function to generate missing data in the feature (x_miss) using the median
        of an observed feature.

        Returns:
            dataset (DataFrame): The dataset with missing values generated under
            the MAR mechanism.

        Reference:
        [1] Santos, M. S., R. C. Pereira, A. F. Costa, J. P. Soares, J. Santos, and
        P. H. Abreu. 2019. Generating Synthetic Missing Data: A Review by Missing Mechanism.
        IEEE Access 7: 11651–67.

        """
        median_xobs = self.dataset[self.x_obs].median()
        pb = np.zeros(self.dataset.shape[0])

        grupos = np.where(self.dataset[self.x_obs] >= median_xobs, 1, 0)

        if len(pb[grupos == 0]) == 0 or len(pb[grupos == 1]) == 0:
            grupos = np.where(self.dataset[self.x_obs] <= median_xobs, 1, 0)

        # The group with values higher or equal to the median have 9 times more probability to be chosen
        pb[grupos == 0] = 0.1 / np.sum(grupos == 0)
        pb[grupos == 1] = 0.9 / np.sum(grupos == 1)
        pos_xmis = np.random.choice(
            np.arange(self.dataset.shape[0]), size=self.N, replace=False, p=pb
        )
        self.dataset.loc[pos_xmis, self.x_miss] = np.nan
        return self.dataset

    # ------------------------------------------------------------------------
    def highest(self):
        """Function to generate missing values in the feature (x_miss) using the
        highest values from an observed feature.

        Returns:
            dataset (DataFrame): The dataset with missing values generated under
            the MAR mechanism.

        Reference:
        [1] Santos, M. S., R. C. Pereira, A. F. Costa, J. P. Soares, J. Santos, and
        P. H. Abreu. 2019. Generating Synthetic Missing Data: A Review by Missing Mechanism.
        IEEE Access 7: 11651–67.

        """

        pos_xmis = (
            self.dataset[self.x_obs]
            .sort_values(ascending=False)[: self.N]
            .index
        )
        self.dataset.loc[pos_xmis, self.x_miss] = np.nan
        return self.dataset

    # ------------------------------------------------------------------------
    def mix(self):
        """
        Function to generate missing values in the feature (x_miss) using the
        N/2 lowest values and N/2 highest values from an observed feature.

        Returns:
            dataset (DataFrame): The dataset with missing values generated under
            the MAR mechanism.

        Reference:
        [1] Santos, M. S., R. C. Pereira, A. F. Costa, J. P. Soares, J. Santos, and
        P. H. Abreu. 2019. Generating Synthetic Missing Data: A Review by Missing Mechanism.
        IEEE Access 7: 11651–67.

        """
        pos_xmis_lowest = (
            self.dataset[self.x_obs].sort_values()[: round(self.N / 2)].index
        )
        pos_xmis_highest = (
            self.dataset[self.x_obs]
            .sort_values(ascending=False)[: round(self.N / 2)]
            .index
        )
        pos_xmis = list(pos_xmis_lowest) + list(pos_xmis_highest)
        self.dataset.loc[pos_xmis, self.x_miss] = np.nan
        return self.dataset



# -*- coding: utf-8 -*

#  =============================================================================
# Aeronautics Institute of Technologies (ITA) - Brazil
# University of Coimbra (UC) - Portugal
# Arthur Dantas Mangussi - mangussiarthur@gmail.com
# =============================================================================

import numpy as np
import pandas as pd
from mdatagen.utils.feature_choice import FeatureChoice
from mdatagen.utils.math_calcs import MathCalcs


# ==========================================================================
class mMAR:
    """
    A class to generate missing data in a dataset based on the Missing At Random (MAR) mechanism for multiple features simultaneously.

    Args:
        X (pd.DataFrame): The dataset to receive the missing data.
        y (np.array): The label values from dataset
        n_xmiss (int): The number of features in the dataset that will receive missing values. Default is 2.
        missTarget (bool, optional): A flag to generate missing into the target.

    Example Usage:

    ```python
    # Create an instance of the MAR class
    generator = MAR(X, y, n_xmiss=4)

    # Generate missing values using the random strategy
    data_md = generator.random(missing_rate = 20)
    ```
    """

    def __init__(self, X: pd.DataFrame, y: np.array, n_xmiss: int = 2, missTarget:bool=False):
        if not isinstance(X, pd.DataFrame):
            raise TypeError('Dataset must be a Pandas Dataframe')
        if not isinstance(y, np.ndarray):
            raise TypeError('y must be a numpy array')

        self.n_xmiss = n_xmiss
        self.X = X
        self.y = y
        self.dataset = self.X.copy()
        self.missTarget = missTarget

        if missTarget:
            self.dataset['target'] = y

    def random(self, missing_rate: int = 10) -> pd.DataFrame:
        """
        Function to generate arficial missing data in n_xmiss features chosen randomly.
        The lower values in observed feature for each feature x_miss will determine
        the miss locations in x_miss.

        Args:
            missing_rate (int, optional): The rate of missing data to be generated. Default is 10.

        Returns:
            dataset (DataFrame): The dataset with missing values generated under
            the MAR mechanism.

        Reference:
        [1] Santos, M. S., R. C. Pereira, A. F. Costa, J. P. Soares, J. Santos, and
        P. H. Abreu. 2019. Generating Synthetic Missing Data: A Review by Missing Mechanism.
        IEEE Access 7: 11651–67.

        """
        if missing_rate >= 100:
            raise ValueError(
                'Missing Rate is too high, the whole feature will be deleted!'
            )
        elif missing_rate < 0:
            raise ValueError('Missing rate must be between [0,100]')

        n = self.dataset.shape[0]
        p = self.dataset.shape[1]
        mr = missing_rate / 100

        N = round((n * p * mr) / self.n_xmiss)

        if self.n_xmiss > p:
            raise ValueError(
                'n_xmiss must be lower than number of features in dataset'
            )
        elif N > n:
            raise ValueError(
                f'FEATURES will be all NaN ({N}>{n}), you should increase the number of features that will receive missing data'
            )

        xmiss_multiva = []
        cont = 0

        while cont < self.n_xmiss:
            x_obs = np.random.choice(self.dataset.columns, replace=False)
            x_miss = np.random.choice(
                [x for x in self.dataset.columns if x != x_obs]
            )

            if x_miss not in xmiss_multiva:
                pos_xmiss = self.dataset[x_obs].sort_values()[:N].index
                self.dataset.loc[pos_xmiss, x_miss] = np.nan

                xmiss_multiva.append(x_miss)
                cont += 1

        if not self.missTarget:
            self.dataset['target'] = self.y
        return self.dataset

    def correlated(self, missing_rate: int = 10) -> pd.DataFrame:
        """
        Function to generate missing data in features from dataset, except the class (target).
        The lower values in observed feature for each correlated pair will determine the
        miss locations in feature x_miss.

        Args:
            missing_rate (int, optional): The rate of missing data to be generated. Default is 10.

        Returns:
            dataset (DataFrame): The dataset with missing values generated under
            the MAR mechanism.

        Reference:
        [1] Santos, M. S., R. C. Pereira, A. F. Costa, J. P. Soares, J. Santos, and
        P. H. Abreu. 2019. Generating Synthetic Missing Data: A Review by Missing Mechanism.
        IEEE Access 7: 11651–67.

        """

        if missing_rate >= 50:
            raise ValueError(
                'Features will be all NaN, you should decrease the missing rate'
            )

        mr = missing_rate / 100

        pairs, correlation_matrix = FeatureChoice._make_pairs(self.X, self.y, self.missTarget)

        for pair in pairs:
            if len(pair) % 2 == 0:
                cutK = 2 * mr
                x_miss = FeatureChoice._find_most_correlated_feature_even(
                    correlation_matrix, pair
                )
                x_obs = next(elem for elem in pair if elem != x_miss)

            else:
                x_miss = FeatureChoice._find_most_correlated_feature_odd(
                   correlation_matrix, pair
                )
                x_obs = set(pair).difference(x_miss).pop()
                cutK = 1.5 * mr

            N = round(len(self.dataset) * cutK)
            pos_xmiss = self.dataset[x_obs].sort_values()[:N].index
            self.dataset.loc[pos_xmiss, x_miss] = np.nan

        if not self.missTarget:
            self.dataset['target'] = self.y
        return self.dataset

    def median(self, missing_rate: int = 10) -> pd.DataFrame:
        """
        Function to generate missing data in features from dataset.
        The median in observed feature for each correlated pair will create two groups.
        The group is chosen randomly, and lower values will determine the miss locations
        in feature x_miss.

        Args:
            missing_rate (int, optional): The rate of missing data to be generated. Default is 10.

        Returns:
            dataset (DataFrame): The dataset with missing values generated under
            the MAR mechanism.

        Reference:
        [1] Santos, M. S., R. C. Pereira, A. F. Costa, J. P. Soares, J. Santos, and
        P. H. Abreu. 2019. Generating Synthetic Missing Data: A Review by Missing Mechanism.
        IEEE Access 7: 11651–67.

        """

        if missing_rate >= 50:
            raise ValueError(
                'Features will be all NaN, you should decrease the missing rate'
            )

        mr = missing_rate / 100

        pairs,_ = FeatureChoice._make_pairs(self.X, self.y, self.missTarget)

        for pair in pairs:
            x_obs = np.random.choice(pair)

            if len(pair) % 2 == 0:
                cutK = 2 * mr
                x_miss = next(elem for elem in pair if elem != x_obs)
            else:
                cutK = 1.5 * mr
                x_miss = list(pair)
                x_miss.remove(x_obs)

            N = round(len(self.dataset) * cutK)

            if len(self.dataset[x_obs].unique()) == 2:  # Binary feature
                g1, g2 = MathCalcs.define_groups(self.dataset, x_obs)

            else:  # Continuos or ordinal feature
                median_xobs = self.dataset[x_obs].median()

                g1 = self.dataset[x_obs][self.dataset[x_obs] <= median_xobs]
                g2 = self.dataset[x_obs][self.dataset[x_obs] > median_xobs]

                if len(g1) != len(
                    g2
                ):  # If median do not divide in equal-size groups
                    g1, g2 = MathCalcs.define_groups(self.dataset, x_obs)

            choice = np.random.choice([0, 1])
            if choice == 0:
                pos_xmiss = g1.sort_values()[:N].index

            else:
                pos_xmiss = g2.sort_values()[:N].index

            self.dataset.loc[pos_xmiss, x_miss] = np.nan

        if not self.missTarget:
            self.dataset['target'] = self.y
        return self.dataset


# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name='mdatagen',
    version='0.1.66',
    keywords=['machine learning', 'preprocessing data'],
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    license='MIT',
    author='Arthur Dantas Mangussi',
    author_email='mangussiarthur@gmail.com',
    url='https://github.com/ArthurMangussi/pymdatagen',
    description='mdatagen: A Python Library for the Generation of Artificial Missing Data',
    long_description=long_description,
    long_description_content_type='text/markdown',

    python_requires='>=3.10.12', 
    install_requires=[
        'numpy >= 1.25.0',
        'pandas >= 2.0.3',
        'scikit-learn >= 1.3.0',
        'missingno >= 0.5.2',
        'scipy >= 1.11.4']
)


# -*- coding: utf-8 -*-

# =============================================================================
# Aeronautics Institute of Technologies (ITA) - Brazil
# University of Coimbra (UC) - Portugal
# Arthur Dantas Mangussi - mangussiarthur@gmail.com
# =============================================================================

import numpy as np
import pandas as pd
from mdatagen.utils.math_calcs import MathCalcs


# ==========================================================================
class uMCAR:
    """
    A class to generate missing values in a dataset based on the Missing Completely At Random (MCAR) univariate mechanism.

    Args:
        X (pd.DataFrame): The dataset to receive the missing data.
        y (np.array): The label values from dataset
        missing_rate (int, optional): The rate of missing data to be generated. Default is 10.
        x_miss (string, optional): The name of feature to insert the missing data.
        method (str, optional): The method to choose x_miss. If x_miss not informed by user, x_miss will be choose randomly. The options to choose xmiss is ["random", "correlated", "min", "max"]. Default is "random"
        seed (int, optional): The seed for the random number generator.


    Example Usage:
    ```python
    # Create an instance of the MCAR class
    generator = MCAR(X, y, missing_rate=20, method="correlated")

    # Generate missing values using the random strategy
    data_md = generator.random()
    ```
    """

    # ------------------------------------------------------------------------
    def __init__(
        self,
        X: pd.DataFrame,
        y: np.array,
        missing_rate: int = 10,
        x_miss: str = None,
        method: str = 'random',
        seed: int = None
    ):
        if not isinstance(X, pd.DataFrame):
            raise TypeError('Dataset must be a Pandas Dataframe')
        if not isinstance(y, np.ndarray):
            raise TypeError('y must be a numpy array')

        self.missing_rate = missing_rate

        if self.missing_rate >= 100:
            raise ValueError(
                'Missing Rate is too high, the whole feature will be deleted!'
            )
        elif self.missing_rate < 0:
            raise ValueError('Missing rate must be between [0,100]')

        self.X = X
        self.y = y
        self.dataset = self.X.copy()
        self.dataset['target'] = y
        self.seed = seed
        if seed is not None:
            np.random.seed(seed)

        self.N = round(self.missing_rate * self.dataset.shape[0] / 100)

        relevance = MathCalcs._nmi(self.X, self.y)

        option_mapping = {
            'random': lambda: np.random.choice(
                list(filter(lambda x: x != 'target', self.dataset.columns))
            ),
            'correlated': lambda: MathCalcs._find_correlation(
                self.X, self.y, 'target'
            ),
            'min': lambda: min(relevance)[1],
            'max': lambda: max(relevance)[1],
        }

        if not x_miss:
            self.x_miss = option_mapping[method]()
        else:
            self.x_miss = x_miss

    # ------------------------------------------------------------------------
    def random(self):
        """
        Function to randomly select locations in the feature (x_miss) to be missing.

        Returns:
            dataset (DataFrame): The dataset with missing values generated under
            the MCAR mechanism.

        Reference:
        [1] Santos, M. S., R. C. Pereira, A. F. Costa, J. P. Soares, J. Santos, and
        P. H. Abreu. 2019. Generating Synthetic Missing Data: A Review by Missing Mechanism.
        IEEE Access 7: 11651–67.

        """
        pos_xmiss = np.random.choice(
            self.dataset[self.x_miss].index, self.N, replace=False
        )
        self.dataset.loc[pos_xmiss, self.x_miss] = np.nan
        return self.dataset

    # ------------------------------------------------------------------------
    def binomial(self):
        """
        Function to choose the feature (x_miss) locations to be missing by Bernoulli distribution.

        Returns:
            dataset (DataFrame): The dataset with missing values generated under
            the MCAR mechanism.

        Reference:
        [1] Santos, M. S., R. C. Pereira, A. F. Costa, J. P. Soares, J. Santos, and
        P. H. Abreu. 2019. Generating Synthetic Missing Data: A Review by Missing Mechanism.
        IEEE Access 7: 11651–67.

        """
        pos_xmiss = np.random.binomial(
            n=1, p=self.missing_rate / 100, size=self.dataset.shape[0]
        ).astype(bool)
        self.dataset.loc[pos_xmiss, self.x_miss] = np.nan
        return self.dataset



# -*- coding: utf-8 -*-

# =============================================================================
# Aeronautics Institute of Technologies (ITA) - Brazil
# University of Coimbra (UC) - Portugal
# Arthur Dantas Mangussi - mangussiarthur@gmail.com
# =============================================================================

import numpy as np
import pandas as pd
from mdatagen.utils.feature_choice import FeatureChoice

# ==========================================================================
class uMNAR:
    """
    A class to generate missing values in a dataset based on the Missing Not At Random (MNAR) univariate mechanism.

    Args:
        X (pd.DataFrame): The dataset to receive the missing data.
        y (np.array): The label values from dataset
        missing_rate (int, optional): The rate of missing data to be generated. Default is 10.
        x_miss (string, optional): The name of feature to insert the missing data. If not informed, x_miss will be the feature most correlated with target
        threshold (float, optional): The threshold to select the locations in feature (xmiss) to receive missing values where 0 indicates de lowest and 1 highest values. Default is 0

    Example Usage:
    ```python
    # Create an instance of the MNAR class
    generator = MNAR(X, y, missing_rate=20, x_miss='feature1', threshold = 0.5)

    # Generate missing values using the lowest strategy
    data_md = generator.run()
    ```
    """

    # ------------------------------------------------------------------------
    def __init__(
        self,
        X: pd.DataFrame,
        y: np.array,
        missing_rate: int = 10,
        x_miss: str = None,
        threshold: float = 0,
    ):
        if not isinstance(X, pd.DataFrame):
            raise TypeError('Dataset must be a Pandas Dataframe')
        if not isinstance(y, np.ndarray):
            raise TypeError('y must be a numpy array')

        self.missing_rate = missing_rate

        if self.missing_rate >= 100:
            raise ValueError(
                'Missing Rate is too high, the whole feature will be deleted!'
            )
        elif self.missing_rate < 0:
            raise ValueError('Missing rate must be between [0,100]')

        self.X = X
        self.y = y
        self.dataset = self.X.copy()
        self.dataset['target'] = y
        self.p = threshold
        self.N = round(self.missing_rate * self.dataset.shape[0] / 100)

        if not (0 <= self.p <= 1):
            raise ValueError('Threshold must be in range [0,1]')

        self.x_miss, self.x_obs = FeatureChoice._define_xmiss(
            self.X, self.y, x_miss, flag=False
        )

    # ------------------------------------------------------------------------
    def run(self, deterministic:bool = False):
        """
        Function to generate missing values in the feature (x_miss) using the
        threshold to choose values from an unobserved feature or feature x_miss
        itself.

        Args:
            deterministc (bool, optinal): A flag that determine if x_miss will have miss 
            locations based on itself or an unobserved feature. Default is False
            (i.e., an unobserved feature).

        Returns:
            dataset (DataFrame): The dataset with missing values generated under the MNAR mechanism.

        Reference:
        [1] Santos, M. S., R. C. Pereira, A. F. Costa, J. P. Soares, J. Santos, and
        P. H. Abreu. 2019. Generating Synthetic Missing Data: A Review by Missing Mechanism.
        IEEE Access 7: 11651–67.

        """

        x_f = self.dataset.loc[:, self.x_miss]

        if deterministic:
            # Observed feature
            ordered_id = x_f.sort_values()
            pos_xmiss = FeatureChoice.miss_locations(ordered_id, self.p, self.N)

        else:
            # Unobserved random feature
            ordered_id = np.lexsort((np.random.random(x_f.size), x_f))
            pos_xmiss = FeatureChoice.miss_locations(ordered_id, self.p, self.N)

        self.dataset.loc[pos_xmiss, self.x_miss] = np.nan

        return self.dataset
